local AceAddon = LibStub("AceAddon-3.0");
local AceDB = LibStub("AceDB-3.0");

ZThreatMeter = AceAddon:NewAddon("ZThreatMeter", "AceConsole-3.0", "AceEvent-3.0");

local _G = getfenv(0);
local UnitName = _G.UnitName;
local UnitClass = _G.UnitClass;
local UnitIsDead = _G.UnitIsDead;
local UnitCanAttack = _G.UnitCanAttack;
local UnitIsPlayer = _G.UnitIsPlayer;
local UnitGUID = _G.UnitGUID;
local UnitThreatSituation = _G.UnitThreatSituation;
local UnitDetailedThreatSituation = _G.UnitDetailedThreatSituation;
local GetThreatStatusColor = _G.GetThreatStatusColor;
local InCombatLockdown = _G.InCombatLockdown;

local math = _G.math;
local table = _G.table;
local RAID_CLASS_COLORS = _G.RAID_CLASS_COLORS;

local members = 0;
local prefix;
local currentTarget = "";
local threatlist = {};
local class = {};

function ZThreatMeter:OnInitialize()
	self.db = AceDB:New("ZThreatMeterDB");

	self.db:RegisterDefaults({
		profile = {
			frame = {
                posx = 0,
                posy = 0,
                scale = 1.0,
				solo = true,
				raid = true,
				party = true,
			},
			tvalue = "sP",
			limit = 5,
			showself = true,
			colorself = true,
			showpets = false,
			showpetssolo = false,
		}
	});

	self:CreateFrame();

	self.name = UnitName("player");
	self.pos = 1;

	self.bars = {};

	RAID_CLASS_COLORS["PET"] = {r = 1, g = 1, b = 0, a = 0.99};
end

function ZThreatMeter:OnEnable()
	self:RegisterEvent("UNIT_THREAT_LIST_UPDATE");
	self:RegisterEvent("PLAYER_TARGET_CHANGED");
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "CheckParty");
	self:RegisterEvent("RAID_ROSTER_UPDATE", "CheckParty");
	self:RegisterEvent("PLAYER_REGEN_ENABLED");

	self:CheckParty();
end

function ZThreatMeter:OnDisable()
	self:UnregisterAllEvents();
end

function ZThreatMeter:PLAYER_TARGET_CHANGED()
	if (not UnitIsDead("target") and UnitCanAttack("player", "target") and not UnitIsPlayer("target")) then
		currentTarget = UnitGUID("target");
		self.title:SetText(UnitName("target"));
		ZThreatMeter:GetThreat("target");
	else
		self.title:SetText("ZThreatMeter");
		self:ClearTarget();
	end
end

function ZThreatMeter:UNIT_THREAT_LIST_UPDATE(event, mob)
	if (mob and UnitGUID(mob) == currentTarget) then
		ZThreatMeter:GetThreat(mob);
	end
end

function ZThreatMeter:OnUpdate()
	if (InCombatLockdown()) then
        local state = UnitThreatSituation("player");
        if (state) then
            local r, g, b = GetThreatStatusColor(state);
            self.frame:SetBackdropColor(r, g, b, 1);
        end
    end
end

function ZThreatMeter:CheckParty(event)
	local p = "raid";
	local m = 0;

	m = GetNumRaidMembers();

	if (m == 0) then
		m = GetNumPartyMembers();
		p = "party";
	end

	if (m > 0 and p == "party") then
		m = m + 1;
	elseif (m == 0) then
		m = 1;
		p = nil;
	end

	prefix = p;
	members = m;

	self:SetFrameVisibility()
end

function ZThreatMeter:PLAYER_REGEN_ENABLED()
	if (self.frame) then
		self.frame:SetBackdropColor(0.69, 0.69, 0.69, 0.99);
	end
end

function ZThreatMeter:GetThreat(mob)
	threatlist = {};

    for i = 1, members do
    	local unit;
		local pet;

    	if (prefix) then
    		if (prefix == "party" and i == members) then
    			unit = "player";
    		else
    			unit = prefix..i;
    		end
    		pet = prefix.."pet"..i;
    	else
    		unit = "player";
	    	pet = "pet";
    	end

    	local _, _, sP, rP, tV = UnitDetailedThreatSituation(unit, mob);
		if (tV) then
			self:ThreatListAdd(unit, sP, rP, tV, nil);
		end

		if (pet and self.db.profile.showpets and ((prefix and not self.db.profile.showpetssolo) or not prefix)) then
            if (UnitExists(pet)) then
                local _, _, sP, rP, tV = UnitDetailedThreatSituation(pet, mob);
                if (tV) then
                    self:ThreatListAdd(pet, sP, rP, tV, UnitName(unit));
                end
            end
		end
    end
	self:UpdateBars();
end

function ZThreatMeter:ThreatListAdd(unit, sP, rP, tV, owner)
        local ins;
        
        for i = 1, #threatlist do
            if (tV >= threatlist[i].tV) then
                ins = i;
                break;
            else
            	ins = i + 1;
            end
        end

        local name = UnitName(unit);

        if (not class[name]) then
        	if (owner) then
        		class[name] = "PET";
            else
            	class[name] = select(2, UnitClass(unit));
           	end
        end

        if (name == self.name) then
            self.pos = ins or 1;
        end

        table.insert(threatlist, ins or 1, {name = name, class = class[name], owner = owner, sP = sP, rP = rP, tV = tV});
end

function ZThreatMeter:UpdateBars()
	local cnt = 0;
	local val = self.db.profile.tvalue;
	local limit = self.db.profile.limit;
	
    for i = 1, #threatlist do
    	local pos = i;

    	if (self.db.profile.showself and i == limit and tonumber(self.pos) > limit) then
			pos = self.pos;
    	end

        if (not self.bars[i]) then
            self:CreateBar(i);
        end

		local c;

		if (self.db.profile.colorself and threatlist[pos].name == self.name) then
			c = {r = 1, g = 0, b = 0, a = 0.99};
		else
	        c =  RAID_CLASS_COLORS[threatlist[pos].class]
		end

		local name, skip;

		if (threatlist[pos].owner) then
            if (self.db.profile.showpets) then
            	name = threatlist[pos].name.." ("..threatlist[pos].owner..")";
            else
            	skip = true;
            end
		else
			name = threatlist[pos].name;
		end

        if (not skip) then
            local str = ("%s   %d (%d%%)"):format(name, math.ceil(tonumber(threatlist[pos].tV)/100), tonumber(threatlist[pos][val]));

            self.bars[i].pos:SetText(tostring(pos)..".");
            self.bars[i].text:SetText(str);
            self.bars[i].frame:SetValue(tonumber(threatlist[pos][val]));
            self.bars[i].frame:SetStatusBarColor(c.r, c.g, c.b, c.a);
            self.bars[i].frame:Show();

            cnt = cnt + 1;

	        if (i >= limit) then break end
	    end
    end

    if (cnt < #self.bars) then
    	cnt = cnt + 1;
    	for i = cnt, #self.bars do
    		self.bars[i].frame:Hide();
    	end
    end
end

function ZThreatMeter:ClearTarget()
	currentTarget = "";
	for i = 1, #self.bars do
		self.bars[i].frame:Hide();
	end
end

function ZThreatMeter:Print(...)
	local str;
	local l = select('#', ...);
	
	str = tostring(select(1, ...));	
	
	if (l > 1) then
        for i = 2, l  do
            str = str..", "..tostring(select(i, ...));
        end
	end
		
	DEFAULT_CHAT_FRAME:AddMessage(str);
end

function ZThreatMeter:CreateBar(num)
	local anchor;
	local y;

	if (self.bars[num-1]) then
		anchor = self.bars[num-1].frame;
		y = -2;
	else
		y = -8;
	end

	local bar = CreateFrame("StatusBar", "ZThreatMeterFrameBar"..num, self.frame, "TextStatusBar");
	bar:SetOrientation("HORIZONTAL");
	bar:SetWidth((self.frame:GetWidth() - 10));
	bar:SetHeight(15);
	bar:SetStatusBarTexture("Interface\\TargetingFrame\\UI-StatusBar");
    bar:SetStatusBarColor(0, 0, 1, 1);
	bar:SetMinMaxValues(0, 100);
	bar:SetValue(0);
    bar:SetPoint("TOP", anchor or self.title, "BOTTOM", 0, y)

	local pos = bar:CreateFontString(nil, "OVERLAY");
    pos:SetFontObject(GameFontHighlightSmall)
    pos:SetJustifyH("LEFT")
    pos:SetJustifyV("TOP")
    pos:SetWidth(20)
    pos:SetText("");
    pos:ClearAllPoints()
    pos:SetPoint("TOP", bar, "LEFT", 11, 6)
    pos:Show()

    local text = bar:CreateFontString(nil, "OVERLAY");
    text:SetFontObject(GameFontHighlightSmall)
    text:SetJustifyH("CENTER")
    text:SetJustifyV("TOP")
    text:SetWidth(160)
    text:SetText("");
    text:ClearAllPoints()
    text:SetPoint("TOP", bar, "TOP", 0, -2)
    text:Show()

	self.bars[num] = {};
    self.bars[num].frame = bar;
    self.bars[num].pos = pos;
    self.bars[num].text = text;
end

function ZThreatMeter:CreateFrame()
	if (self.frame) then return end

	local frame = CreateFrame("Frame", "ZThreatMeterFrame", UIParent);
	  	  frame:EnableMouse(true)
		  frame:SetMovable(true)
		  frame:RegisterForDrag("LeftButton")
		  frame:SetScript("OnDragStart", function() if IsAltKeyDown() then self.frame:StartMoving() end end)
		  frame:SetScript("OnDragStop", function() self.frame:StopMovingOrSizing() self:SavePosition() end)
		  frame:SetScript("OnMouseUp", function (f, b) if (b == "RightButton" and IsAltKeyDown()) then self:ShowConfig() end end)
		  frame:SetScript("OnUpdate", function() self:OnUpdate() end);
		  frame:SetScript("OnHide", function () self:OnDisable() end);
		  frame:SetScript("OnShow", function () self:OnEnable() end);
		  frame:SetWidth(200)
		  frame:SetHeight(24)
		  frame:SetScale(self.db.profile.frame.scale);
		  frame:SetBackdrop({bgFile = "Interface/Tooltips/UI-Tooltip-Background",
                                            edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
                                            tile = true, tileSize = 16, edgeSize = 16,
                                            insets = { left = 4, right = 4, top = 4, bottom = 4 }});
		  frame:SetBackdropColor(0.69, 0.69, 0.69, 0.99);
		  frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
		  frame:Hide();

    local title = frame:CreateFontString(nil, "ARTWORK")
          title:SetFontObject(GameFontNormalSmall)
          title:SetText("ZThreatMeter")
          title:SetJustifyH("CENTER")
          title:SetWidth(160)
          title:SetHeight(12)
          title:Show()
          title:ClearAllPoints()
          title:SetPoint("TOP", frame, "TOP", 0, -5)

	self.frame = frame;
	self.title = title;
end

function ZThreatMeter:SavePosition()
	local f = self.frame
	local x,y = f:GetLeft(), f:GetTop()
	local s = f:GetEffectiveScale()

	x,y = x*s,y*s

	self.db.profile.frame.posx = x
	self.db.profile.frame.posy = y
end

function ZThreatMeter:RestorePosition()
	local x = self.db.profile.frame.posx
	local y = self.db.profile.frame.posy

	if not x or not y then return end

	local f = self.frame
	local s = f:GetEffectiveScale()

	x,y = x/s,y/s

	f:ClearAllPoints()
	f:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x, y)
end

function ZThreatMeter:UpdateScale()
	if self.frame then
		self.frame:SetScale(self.db.profile.frame.scale)
		self:RestorePosition()
	end
end

function ZThreatMeter:SetFrameVisibility()
	local visible = self.frame:IsVisible();
	local p = prefix or "solo";

    if (self.db.profile.frame[p] and not visible) then
            self.frame:Show();
    elseif (not self.db.profile.frame[p] and visible) then
        	self.frame:Hide();
	end

	--[[
    if (not prefix) then
        if (self.db.profile.frame.solo and not visible) then
            self.frame:Show();
        elseif (not self.db.profile.frame.solo and visible) then
        	self.frame:Hide();
		end
    elseif (prefix == "party") then
        if (self.db.profile.frame.party and not visible) then
            self.frame:Show();
        elseif (not self.db.profile.frame.party and visible) then
        	self.frame:Hide();
        end
    elseif (prefix == "raid") then
        if (self.db.profile.frame.raid and not visible) then
            self.frame:Show();
        elseif (not self.db.profile.frame.raid and visible) then
        	self.frame:Hide();
		end
    end
    ]]--
end